// Copyright 2022 the V8 project authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// Flags: --allow-natives-syntax --turbofan --no-always-turbofan

(function optimize() {
  function f() {
    return 'abc'.includes('a');
  }
  v8PrepareFunctionForOptimization(f);
  assertEquals(true, f());
  assertEquals(true, f());
  assertEquals(true, f());
  v8OptimizeFunctionOnNextCall(f);
  assertEquals(true, f());
  assertTrue(isOptimized(f));

  function f2() {
    return 'abc'.includes('a', 1);
  }
  v8PrepareFunctionForOptimization(f2);
  assertEquals(false, f2());
  assertEquals(false, f2());
  assertEquals(false, f2());
  v8OptimizeFunctionOnNextCall(f2);
  assertEquals(false, f2());
  assertTrue(isOptimized(f2));

  function f3() {
    return 'abc'.includes('b');
  }
  v8PrepareFunctionForOptimization(f3);
  assertEquals(true, f3());
  assertEquals(true, f3());
  assertEquals(true, f3());
  v8OptimizeFunctionOnNextCall(f3);
  assertEquals(true, f3());
  assertTrue(isOptimized(f3));

  function f4() {
    return 'abcbc'.includes('bc', 2);
  }
  v8PrepareFunctionForOptimization(f4);
  assertEquals(true, f4());
  assertEquals(true, f4());
  assertEquals(true, f4());
  v8OptimizeFunctionOnNextCall(f4);
  assertEquals(true, f4());
  assertTrue(isOptimized(f4));

  function f5() {
    return 'abcbc'.includes('b', -1);
  }
  v8PrepareFunctionForOptimization(f5);
  assertEquals(true, f5());
  assertEquals(true, f5());
  assertEquals(true, f5());
  v8OptimizeFunctionOnNextCall(f5);
  assertEquals(true, f5());
  assertTrue(isOptimized(f5));

  function f6() {
    return 'abcbc'.includes('b', -10737418);
  }
  v8PrepareFunctionForOptimization(f6);
  assertEquals(true, f6());
  assertEquals(true, f6());
  assertEquals(true, f6());
  v8OptimizeFunctionOnNextCall(f6);
  assertEquals(true, f6());
  assertTrue(isOptimized(f6));
})();

(function optimizeOSR() {
  function f() {
    var result;
    for (var i = 0; i < 100000; i++) {
      result = 'abc'.includes('a');
    }
    return result;
  }
  assertEquals(true, f());

  function f2() {
    var result;
    for (var i = 0; i < 100000; i++) {
      result = 'abc'.includes('a', 1);
    }
    return result;
  }
  assertEquals(false, f2());

  function f3() {
    var result;
    for (var i = 0; i < 100000; i++) {
      result = 'abc'.includes('b');
    }
    return result;
  }
  assertEquals(true, f3());

  function f4() {
    var result;
    for (var i = 0; i < 100000; i++) {
      result = 'abcbc'.includes('bc', 2);
    }
    return result;
  }
  assertEquals(true, f4());
})();

(function bailout() {
  function f(str) {
    return String.prototype.includes.call(str, 'a')
  }
  v8PrepareFunctionForOptimization(f);
  assertEquals(true, f('abc'));
  v8OptimizeFunctionOnNextCall(f);
  assertEquals(true, f({
                 toString: () => {
                   return 'abc'
                 }
  }));
  assertFalse(isOptimized(f));

  function f2(str) {
    return 'abc'.includes(str)
  }
  v8PrepareFunctionForOptimization(f2);
  assertEquals(true, f2('a'));
  v8OptimizeFunctionOnNextCall(f2);
  assertEquals(true, f2({
                 toString: () => {
                   return 'a'
                 }
               }));
  assertFalse(isOptimized(f2));

  function f3(index) {
    return 'abc'.includes('a', index)
  }
  v8PrepareFunctionForOptimization(f3);
  assertEquals(true, f3(0));
  v8OptimizeFunctionOnNextCall(f3);
  assertEquals(true, f3({
                 valueOf: () => {
                   return 0
                 }
  }));
  assertFalse(isOptimized(f3));
})();
